@@ -56,8 +56,9 @@ from .subscription import (
     DEFAULT_SUB_PENDING_MSGS_LIMIT,
     Subscription,
 )
+from .transport import Transport, TcpTransport, WebSocketTransport
 
-__version__ = '2.1.7'
+__version__ = '2.2.0'
 __lang__ = 'python3'
 _logger = logging.getLogger(__name__)
 PROTOCOL = 1
@@ -115,6 +116,61 @@ class Srv:
     did_connect: bool = False
     discovered: bool = False
     tls_name: Optional[str] = None
+    server_version: Optional[str] = None
+
+
+class ServerVersion:
+
+    def __init__(self, server_version: str):
+        self._server_version = server_version
+        self._major_version = None
+        self._minor_version = None
+        self._patch_version = None
+        self._dev_version = None
+
+    def parse_version(self):
+        v = (self._server_version).split('-')
+        if len(v) > 1:
+            self._dev_version = v[1]
+        tokens = v[0].split('.')
+        n = len(tokens)
+        if n > 1:
+            self._major_version = int(tokens[0])
+        if n > 2:
+            self._minor_version = int(tokens[1])
+        if n > 3:
+            self._patch_version = int(tokens[2])
+
+    @property
+    def major(self) -> int:
+        version = self._major_version
+        if not version:
+            self.parse_version()
+        return self._major_version
+
+    @property
+    def minor(self) -> int:
+        version = self._minor_version
+        if not version:
+            self.parse_version()
+        return self._minor_version
+
+    @property
+    def patch(self) -> int:
+        version = self._patch_version
+        if not version:
+            self.parse_version()
+        return self._patch_version
+
+    @property
+    def dev(self) -> int:
+        version = self._dev_version
+        if not version:
+            self.parse_version()
+        return self._dev_version
+
+    def __repr__(self) -> str:
+        return f"<nats server v{self._server_version}>"
 
 
 async def _default_error_callback(ex: Exception) -> None:
@@ -153,10 +209,7 @@ class Client:
         self._pings_outstanding: int = 0
         self._pongs_received: int = 0
         self._pongs: List[asyncio.Future] = []
-        self._bare_io_reader: Optional[asyncio.StreamReader] = None
-        self._io_reader: Optional[asyncio.StreamReader] = None
-        self._bare_io_writer: Optional[asyncio.StreamWriter] = None
-        self._io_writer: Optional[asyncio.StreamWriter] = None
+        self._transport: Optional[Transport] = None
         self._err: Optional[Exception] = None
 
         # callbacks
@@ -594,7 +647,7 @@ class Client:
         ):
             self._reconnection_task.cancel()
 
-            # Wait for the reconection task to be done which should be soon.
+            # Wait for the reconnection task to be done which should be soon.
             try:
                 if self._reconnection_task_future is not None and not self._reconnection_task_future.cancelled(
                 ):
@@ -608,14 +661,14 @@ class Client:
         # Relinquish control to allow background tasks to wrap up.
         await asyncio.sleep(0)
 
-        assert self._io_writer, "Client.connect must be called first"
+        assert self._transport, "Client.connect must be called first"
         if self._current_server is not None:
             # In case there is any pending data at this point, flush before disconnecting.
             if self._pending_data_size > 0:
-                self._io_writer.writelines(self._pending[:])
+                self._transport.writelines(self._pending[:])
                 self._pending = []
                 self._pending_data_size = 0
-                await self._io_writer.drain()
+                await self._transport.drain()
 
         # Cleanup subscriptions since not reconnecting so no need
         # to replay the subscriptions anymore.
@@ -623,12 +676,14 @@ class Client:
             # FIXME: Should we clear the pending queue here?
             if sub._wait_for_msgs_task and not sub._wait_for_msgs_task.done():
                 sub._wait_for_msgs_task.cancel()
+            if sub._message_iterator:
+                sub._message_iterator._cancel()
         self._subs.clear()
 
-        if self._io_writer is not None:
-            self._io_writer.close()
+        if self._transport is not None:
+            self._transport.close()
             try:
-                await self._io_writer.wait_closed()
+                await self._transport.wait_closed()
             except Exception as e:
                 await self._error_cb(e)
 
@@ -638,8 +693,9 @@ class Client:
             if self._closed_cb is not None:
                 await self._closed_cb()
 
-        # Set the client_id back to None
+        # Set the client_id and subscription prefix back to None
         self._client_id = None
+        self._resp_sub_prefix = None
 
     async def drain(self) -> None:
         """
@@ -719,7 +775,7 @@ class Client:
                 # Publish with a reply
                 await nc.publish('hello', b'Hello World!', reply=inbox)
 
-                # Publish with a reply
+                # Publish with headers
                 await nc.publish('hello', b'With Headers', headers={'Foo':'Bar'})
 
                 while True:
@@ -1064,7 +1120,7 @@ class Client:
     @property
     def last_error(self) -> Optional[Exception]:
         """
-        Returns the last error which may have occured.
+        Returns the last error which may have occurred.
         """
         return self._err
 
@@ -1099,6 +1155,27 @@ class Client:
     def is_draining_pubs(self) -> bool:
         return self._status == Client.DRAINING_PUBS
 
+    @property
+    def connected_server_version(self) -> ServerVersion:
+        """
+        Returns the ServerVersion of the server to which the client
+        is currently connected.
+        """
+        if self._current_server and self._current_server.server_version:
+            return ServerVersion(self._current_server.server_version)
+        return ServerVersion("0.0.0-unknown")
+
+    @property
+    def ssl_context(self) -> ssl.SSLContext:
+        ssl_context: Optional[ssl.SSLContext] = None
+        if "tls" in self.options:
+            ssl_context = self.options.get('tls')
+        else:
+            ssl_context = ssl.create_default_context()
+        if ssl_context is None:
+            raise errors.Error('nats: no ssl context provided')
+        return ssl_context
+
     async def _send_command(self, cmd: bytes, priority: bool = False) -> None:
         if priority:
             self._pending.insert(0, cmd)
@@ -1140,6 +1217,8 @@ class Client:
                     # Closer to how the Go client handles this.
                     # e.g. nats://localhost:4222
                     uri = urlparse(connect_url)
+                elif "ws://" in connect_url or "wss://" in connect_url:
+                    uri = urlparse(connect_url)
                 elif ":" in connect_url:
                     # Expand the scheme for the user
                     # e.g. localhost:4222
@@ -1166,6 +1245,14 @@ class Client:
                     self._server_pool.append(Srv(uri))
             except ValueError:
                 raise errors.Error("nats: invalid connect url option")
+            # make sure protocols aren't mixed
+            if not (all(server.uri.scheme in ("nats", "tls")
+                        for server in self._server_pool)
+                    or all(server.uri.scheme in ("ws", "wss")
+                           for server in self._server_pool)):
+                raise errors.Error(
+                    "nats: mixing of websocket and non websocket URLs is not allowed"
+                )
         else:
             raise errors.Error("nats: invalid connect url option")
 
@@ -1196,23 +1283,27 @@ class Client:
                 await asyncio.sleep(self.options["reconnect_time_wait"])
             try:
                 s.last_attempt = time.monotonic()
-                connection_future = asyncio.open_connection(
-                    s.uri.hostname, s.uri.port, limit=DEFAULT_BUFFER_SIZE
-                )
-                r, w = await asyncio.wait_for(
-                    connection_future, self.options['connect_timeout']
-                )
+                if not self._transport:
+                    if s.uri.scheme in ("ws", "wss"):
+                        self._transport = WebSocketTransport()
+                    else:
+                        # use TcpTransport as a fallback
+                        self._transport = TcpTransport()
+                if s.uri.scheme == "wss":
+                    # wss is expected to connect directly with tls
+                    await self._transport.connect_tls(
+                        s.uri,
+                        ssl_context=self.ssl_context,
+                        buffer_size=DEFAULT_BUFFER_SIZE,
+                        connect_timeout=self.options['connect_timeout']
+                    )
+                else:
+                    await self._transport.connect(
+                        s.uri,
+                        buffer_size=DEFAULT_BUFFER_SIZE,
+                        connect_timeout=self.options['connect_timeout']
+                    )
                 self._current_server = s
-
-                # We keep a reference to the initial transport we used when
-                # establishing the connection in case we later upgrade to TLS
-                # after getting the first INFO message. This is in order to
-                # prevent the GC closing the socket after we send CONNECT
-                # and replace the transport.
-                #
-                # See https://github.com/nats-io/asyncio-nats/issues/43
-                self._bare_io_reader = self._io_reader = r
-                self._bare_io_writer = self._io_writer = w
                 break
             except Exception as e:
                 s.last_attempt = time.monotonic()
@@ -1255,7 +1346,7 @@ class Client:
 
     async def _process_op_err(self, e: Exception) -> None:
         """
-        Process errors which occured while reading or parsing
+        Process errors which occurred while reading or parsing
         the protocol. If allow_reconnect is enabled it will
         try to switch the server to which it is currently connected
         otherwise it will disconnect.
@@ -1294,10 +1385,10 @@ class Client:
         ):
             self._flusher_task.cancel()
 
-        if self._io_writer is not None:
-            self._io_writer.close()
+        if self._transport is not None:
+            self._transport.close()
             try:
-                await self._io_writer.wait_closed()
+                await self._transport.wait_closed()
             except Exception as e:
                 await self._error_cb(e)
 
@@ -1320,7 +1411,7 @@ class Client:
                 # Try to establish a TCP connection to a server in
                 # the cluster then send CONNECT command to it.
                 await self._select_next_server()
-                assert self._io_writer, "_select_next_server must've set _io_writer"
+                assert self._transport, "_select_next_server must've set _transport"
                 await self._process_connect_init()
 
                 # Consider a reconnect to be done once CONNECT was
@@ -1348,16 +1439,16 @@ class Client:
                     sub_cmd = prot_command.sub_cmd(
                         sub._subject, sub._queue, sid
                     )
-                    self._io_writer.write(sub_cmd)
+                    self._transport.write(sub_cmd)
 
                     if max_msgs > 0:
                         unsub_cmd = prot_command.unsub_cmd(sid, max_msgs)
-                        self._io_writer.write(unsub_cmd)
+                        self._transport.write(unsub_cmd)
 
                 for sid in subs_to_remove:
                     self._subs.pop(sid)
 
-                await self._io_writer.drain()
+                await self._transport.drain()
 
                 # Flush pending data before continuing in connected status.
                 # FIXME: Could use future here and wait for an error result
@@ -1588,7 +1679,7 @@ class Client:
             if hdr:
                 ctrl_msg = self._is_control_message(data, hdr)
 
-                # Check if the hearbeat has a "Consumer Stalled" header, if
+                # Check if the heartbeat has a "Consumer Stalled" header, if
                 # so, the value is the FC reply to send a nil message to.
                 # We will send it at the end of this function.
                 if ctrl_msg and ctrl_msg.startswith("Idle"):
@@ -1752,12 +1843,11 @@ class Client:
         with authentication.  It is also responsible of setting up the
         reading and ping interval tasks from the client.
         """
-        assert self._io_reader, "must be called only from Client.connect"
-        assert self._io_writer, "must be called only from Client.connect"
+        assert self._transport, "must be called only from Client.connect"
         assert self._current_server, "must be called only from Client.connect"
         self._status = Client.CONNECTING
 
-        connection_completed = self._io_reader.readline()
+        connection_completed = self._transport.readline()
         info_line = await asyncio.wait_for(
             connection_completed, self.options["connect_timeout"]
         )
@@ -1776,6 +1866,9 @@ class Client:
         self._server_info = srv_info
         self._process_info(srv_info, initial_connection=True)
 
+        if 'version' in self._server_info:
+            self._current_server.server_version = self._server_info['version']
+
         if 'max_payload' in self._server_info:
             self._max_payload = self._server_info["max_payload"]
 
@@ -1784,14 +1877,6 @@ class Client:
 
         if 'tls_required' in self._server_info and self._server_info[
                 'tls_required']:
-            ssl_context: Optional[ssl.SSLContext] = None
-            if "tls" in self.options:
-                ssl_context = self.options.get('tls')
-            elif self._current_server.uri.scheme == 'tls':
-                ssl_context = ssl.create_default_context()
-            if ssl_context is None:
-                raise errors.Error('nats: no ssl context provided')
-
             # Check whether to reuse the original hostname for an implicit route.
             hostname = None
             if "tls_hostname" in self.options:
@@ -1801,55 +1886,26 @@ class Client:
             else:
                 hostname = self._current_server.uri.hostname
 
-            await self._io_writer.drain()  # just in case something is left
+            await self._transport.drain()  # just in case something is left
 
-            # loop.start_tls was introduced in python 3.7
-            # the previous method is removed in 3.9
-            if sys.version_info.minor >= 7:
-                # manually recreate the stream reader/writer with a tls upgraded transport
-                reader = asyncio.StreamReader()
-                protocol = asyncio.StreamReaderProtocol(reader)
-                transport_future = asyncio.get_running_loop().start_tls(
-                    self._io_writer.transport,
-                    protocol,
-                    ssl_context,
-                    server_hostname=hostname
-                )
-                transport = await asyncio.wait_for(
-                    transport_future, self.options['connect_timeout']
-                )
-                writer = asyncio.StreamWriter(
-                    transport, protocol, reader, asyncio.get_running_loop()
-                )
-                self._io_reader, self._io_writer = reader, writer
-            else:
-                transport = self._io_writer.transport
-                sock = transport.get_extra_info('socket')
-                if not sock:
-                    # This shouldn't happen
-                    raise errors.Error('nats: unable to get socket')
-
-                connection_future = asyncio.open_connection(
-                    limit=DEFAULT_BUFFER_SIZE,
-                    sock=sock,
-                    ssl=ssl_context,
-                    server_hostname=hostname,
-                )
-                self._io_reader, self._io_writer = await asyncio.wait_for(
-                    connection_future, self.options['connect_timeout']
-                )
+            # connect to transport via tls
+            await self._transport.connect_tls(
+                hostname,
+                self.ssl_context,
+                DEFAULT_BUFFER_SIZE,
+                self.options['connect_timeout'],
+            )
 
         # Refresh state of parser upon reconnect.
         if self.is_reconnecting:
             self._ps.reset()
 
-        assert self._io_reader
-        assert self._io_writer
+        assert self._transport
         connect_cmd = self._connect_command()
-        self._io_writer.write(connect_cmd)
-        await self._io_writer.drain()
+        self._transport.write(connect_cmd)
+        await self._transport.drain()
         if self.options["verbose"]:
-            future = self._io_reader.readline()
+            future = self._transport.readline()
             next_op = await asyncio.wait_for(
                 future, self.options["connect_timeout"]
             )
@@ -1865,10 +1921,10 @@ class Client:
                 # await self._process_err(err_msg)
                 raise errors.Error("nats: " + err_msg.rstrip('\r\n'))
 
-        self._io_writer.write(PING_PROTO)
-        await self._io_writer.drain()
+        self._transport.write(PING_PROTO)
+        await self._transport.drain()
 
-        future = self._io_reader.readline()
+        future = self._transport.readline()
         next_op = await asyncio.wait_for(
             future, self.options["connect_timeout"]
         )
@@ -1902,11 +1958,12 @@ class Client:
         )
 
     async def _send_ping(self, future: asyncio.Future = None) -> None:
-        assert self._io_writer, "Client.connect must be called first"
+        assert self._transport, "Client.connect must be called first"
         if future is None:
             future = asyncio.Future()
         self._pongs.append(future)
-        self._io_writer.write(PING_PROTO)
+        self._transport.write(PING_PROTO)
+        self._pending_data_size += len(PING_PROTO)
         await self._flush_pending()
 
     async def _flusher(self) -> None:
@@ -1915,7 +1972,7 @@ class Client:
         and then flushes them to the socket.
         """
         assert self._error_cb, "Client.connect must be called first"
-        assert self._io_writer, "Client.connect must be called first"
+        assert self._transport, "Client.connect must be called first"
         assert self._flush_queue, "Client.connect must be called first"
         while True:
             if not self.is_connected or self.is_connecting:
@@ -1925,10 +1982,10 @@ class Client:
 
             try:
                 if self._pending_data_size > 0:
-                    self._io_writer.writelines(self._pending[:])
+                    self._transport.writelines(self._pending[:])
                     self._pending = []
                     self._pending_data_size = 0
-                    await self._io_writer.drain()
+                    await self._transport.drain()
             except OSError as e:
                 await self._error_cb(e)
                 await self._process_op_err(e)
@@ -1966,15 +2023,15 @@ class Client:
         while True:
             try:
                 should_bail = self.is_closed or self.is_reconnecting
-                if should_bail or self._io_reader is None:
+                if should_bail or self._transport is None:
                     break
-                if self.is_connected and self._io_reader.at_eof():
+                if self.is_connected and self._transport.at_eof():
                     err = errors.UnexpectedEOF()
                     await self._error_cb(err)
                     await self._process_op_err(err)
                     break
 
-                b = await self._io_reader.read(DEFAULT_BUFFER_SIZE)
+                b = await self._transport.read(DEFAULT_BUFFER_SIZE)
                 await self._ps.parse(b)
             except errors.ProtocolError:
                 await self._process_op_err(errors.ProtocolError())
