@@ -11,10 +11,10 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 #
-
 import datetime
+import json
 from dataclasses import dataclass
-from typing import TYPE_CHECKING, Dict, List, Optional
+from typing import TYPE_CHECKING, Dict, List, Optional, Union
 
 from nats.errors import Error, MsgAlreadyAckdError, NotJSMessageError
 
@@ -31,7 +31,6 @@ class Msg:
     subject: str = ''
     reply: str = ''
     data: bytes = b''
-    sid: int = 0
     headers: Optional[Dict[str, str]] = None
 
     _metadata: Optional["Metadata"] = None
@@ -102,12 +101,19 @@ class Msg:
         self._ackd = True
         return resp
 
-    async def nak(self) -> None:
+    async def nak(self, delay: Optional[Union[int, float]] = None) -> None:
         """
         nak negatively acknowledges a message delivered by JetStream triggering a redelivery.
+        if `delay` is provided, redelivery is delayed for `delay` seconds
         """
         self._check_reply()
-        await self._client.publish(self.reply, Msg.Ack.Nak)
+        payload = Msg.Ack.Nak
+        json_args = dict()
+        if delay:
+            json_args['delay'] = int(delay * 10**9)  # to seconds to ns
+        if json_args:
+            payload += (b' ' + json.dumps(json_args).encode())
+        await self._client.publish(self.reply, payload)
         self._ackd = True
 
     async def in_progress(self) -> None:
@@ -141,20 +147,47 @@ class Msg:
         if metadata is not None:
             return metadata
 
-        # TODO: Support V2TokenCount style with domains.
         tokens = Msg.Metadata._get_metadata_fields(msg.reply)
-        t = datetime.datetime.fromtimestamp(int(tokens[7]) / 1_000_000_000.0)
-        metadata = Msg.Metadata(
-            sequence=Msg.Metadata.SequencePair(
-                stream=int(tokens[5]),
-                consumer=int(tokens[6]),
-            ),
-            num_delivered=int(tokens[4]),
-            num_pending=int(tokens[8]),
-            timestamp=t,
-            stream=tokens[2],
-            consumer=tokens[3],
-        )
+
+        if len(tokens) == Msg.Ack.V1TokenCount:
+            t = datetime.datetime.fromtimestamp(
+                int(tokens[7]) / 1_000_000_000.0
+            )
+            metadata = Msg.Metadata(
+                sequence=Msg.Metadata.SequencePair(
+                    stream=int(tokens[5]),
+                    consumer=int(tokens[6]),
+                ),
+                num_delivered=int(tokens[4]),
+                num_pending=int(tokens[8]),
+                timestamp=t,
+                stream=tokens[2],
+                consumer=tokens[3],
+            )
+        else:
+            t = datetime.datetime.fromtimestamp(
+                int(tokens[Msg.Ack.Timestamp]) / 1_000_000_000.0
+            )
+
+            # Underscore indicate no domain is set. Expose as empty string
+            # to client.
+            domain = tokens[Msg.Ack.Domain]
+            if domain == "_":
+                domain = ""
+
+            metadata = Msg.Metadata(
+                sequence=Msg.Metadata.SequencePair(
+                    stream=int(tokens[Msg.Ack.StreamSeq]),
+                    consumer=int(tokens[Msg.Ack.ConsumerSeq]),
+                ),
+                num_delivered=int(tokens[Msg.Ack.NumDelivered]),
+                num_pending=int(tokens[Msg.Ack.NumPending]),
+                timestamp=t,
+                stream=tokens[Msg.Ack.Stream],
+                consumer=tokens[Msg.Ack.Consumer],
+                domain=domain,
+            )
+
         msg._metadata = metadata
         return metadata
 
@@ -187,6 +220,7 @@ class Msg:
         timestamp: Optional[datetime.datetime] = None
         stream: Optional[str] = None
         consumer: Optional[str] = None
+        domain: Optional[str] = None
 
         @dataclass
         class SequencePair:
@@ -198,11 +232,12 @@ class Msg:
 
         @classmethod
         def _get_metadata_fields(cls, reply: Optional[str]) -> List[str]:
-            if reply is None or reply == '':
+            if not reply:
                 raise NotJSMessageError
             tokens = reply.split('.')
-            if len(tokens) != Msg.Ack.V1TokenCount or \
-                    tokens[0] != Msg.Ack.Prefix0 or \
-                    tokens[1] != Msg.Ack.Prefix1:
-                raise NotJSMessageError
-            return tokens
+            if (len(tokens) == Msg.Ack.V1TokenCount or
+                    len(tokens) >= Msg.Ack.V2TokenCount-1) and \
+                    tokens[0] == Msg.Ack.Prefix0 and \
+                    tokens[1] == Msg.Ack.Prefix1:
+                return tokens
+            raise NotJSMessageError
