@@ -1,4 +1,4 @@
-# Copyright 2021 The NATS Authors
+# Copyright 2021-2022 The NATS Authors
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
 # You may obtain a copy of the License at
@@ -13,7 +13,6 @@
 #
 
 import asyncio
-import base64
 import json
 import time
 from typing import TYPE_CHECKING, Awaitable, Callable, List, Optional
@@ -43,6 +42,9 @@ Callback = Callable[['Msg'], Awaitable[None]]
 DEFAULT_JS_SUB_PENDING_MSGS_LIMIT = 512 * 1024
 DEFAULT_JS_SUB_PENDING_BYTES_LIMIT = 256 * 1024 * 1024
 
+# Max history limit for key value.
+KV_MAX_HISTORY = 64
+
 
 class JetStreamContext(JetStreamManager):
     """
@@ -143,6 +145,8 @@ class JetStreamContext(JetStreamManager):
         pending_msgs_limit: Optional[int] = DEFAULT_JS_SUB_PENDING_MSGS_LIMIT,
         pending_bytes_limit: Optional[int
                                       ] = DEFAULT_JS_SUB_PENDING_BYTES_LIMIT,
+        deliver_policy: Optional[api.DeliverPolicy] = None,
+        headers_only: Optional[bool] = None,
     ) -> Subscription:
         """Create consumer if needed and push-subscribe to it.
 
@@ -258,6 +262,11 @@ class JetStreamContext(JetStreamManager):
                 config.durable_name = durable
             if not config.deliver_group:
                 config.deliver_group = queue
+            if not config.headers_only:
+                config.headers_only = headers_only
+            if deliver_policy:
+                # NOTE: deliver_policy is defaulting to ALL so check is different for this one.
+                config.deliver_policy = deliver_policy
 
             # Create inbox for push consumer.
             deliver = self._nc.new_inbox()
@@ -277,10 +286,12 @@ class JetStreamContext(JetStreamManager):
             # one message being delivered at a time.
             if ordered_consumer:
                 config.flow_control = True
-                config.ack_policy = api.AckPolicy.EXPLICIT
+                config.ack_policy = api.AckPolicy.NONE
                 config.max_deliver = 1
                 config.ack_wait = 22 * 3600  # 22 hours
                 config.idle_heartbeat = idle_heartbeat
+                config.num_replicas = 1
+                config.memory_storage = True
 
             consumer_info = await self._jsm.add_consumer(stream, config=config)
             consumer = consumer_info.name
@@ -316,7 +327,11 @@ class JetStreamContext(JetStreamManager):
         """
         # By default, async subscribers wrap the original callback and
         # auto ack the messages as they are delivered.
-        if cb and not manual_ack:
+        #
+        # In case ack policy is none then we also do not require to ack.
+        #
+        if cb and (not manual_ack) and (config.ack_policy
+                                        is not api.AckPolicy.NONE):
             cb = self._auto_ack_callback(cb)
         if config.deliver_subject is None:
             raise TypeError("config.deliver_subject is required")
@@ -627,6 +642,7 @@ class JetStreamContext(JetStreamManager):
             self._stream = stream
             self._consumer = consumer
 
+            self._sub = sub
             self._conn = sub._conn
             self._id = sub._id
             self._subject = sub._subject
@@ -655,6 +671,13 @@ class JetStreamContext(JetStreamManager):
             )
             return info
 
+        @property
+        def delivered(self) -> int:
+            """
+            Number of delivered messages to this subscription so far.
+            """
+            return self._sub._received
+
     class PullSubscription:
         """
         PullSubscription is a subscription that can fetch messages.
@@ -696,6 +719,13 @@ class JetStreamContext(JetStreamManager):
             """
             return self._sub._pending_size
 
+        @property
+        def delivered(self) -> int:
+            """
+            Number of delivered messages to this subscription so far.
+            """
+            return self._sub._received
+
         async def unsubscribe(self) -> None:
             """
             unsubscribe destroys de inboxes of the pull subscription making it
@@ -954,40 +984,6 @@ class JetStreamContext(JetStreamManager):
 
             return msgs
 
-    #############################
-    #                           #
-    # JetStream Manager Context #
-    #                           #
-    #############################
-
-    async def get_last_msg(
-        self,
-        stream_name: str,
-        subject: str,
-    ) -> api.RawStreamMsg:
-        """
-        get_last_msg retrieves a message from a stream.
-        """
-        req_subject = f"{self._prefix}.STREAM.MSG.GET.{stream_name}"
-        req = {'last_by_subj': subject}
-        data = json.dumps(req)
-        resp = await self._api_request(
-            req_subject, data.encode(), timeout=self._timeout
-        )
-        raw_msg = api.RawStreamMsg.from_response(resp['message'])
-        if raw_msg.hdrs:
-            hdrs = base64.b64decode(raw_msg.hdrs)
-            raw_headers = hdrs[NATS_HDR_LINE_SIZE + _CRLF_LEN_:]
-            parsed_headers = self._jsm._hdr_parser.parsebytes(raw_headers)
-            headers = None
-            if len(parsed_headers.items()) > 0:
-                headers = {}
-                for k, v in parsed_headers.items():
-                    headers[k] = v
-            raw_msg.headers = headers
-
-        return raw_msg
-
     ######################
     #                    #
     # KeyValue Context   #
@@ -1008,6 +1004,7 @@ class JetStreamContext(JetStreamManager):
             stream=stream,
             pre=KV_PRE_TEMPLATE.format(bucket=bucket),
             js=self,
+            direct=si.config.allow_direct
         )
 
     async def create_key_value(
@@ -1022,27 +1019,41 @@ class JetStreamContext(JetStreamManager):
             config = api.KeyValueConfig(bucket=params["bucket"])
         config = config.evolve(**params)
 
+        duplicate_window = 2 * 60  # 2 minutes
+        if config.ttl and config.ttl < duplicate_window:
+            duplicate_window = config.ttl
+
+        if config.history > 64:
+            raise nats.js.errors.KeyHistoryTooLargeError
+
         stream = api.StreamConfig(
             name=KV_STREAM_TEMPLATE.format(bucket=config.bucket),
-            description=None,
+            description=config.description,
             subjects=[f"$KV.{config.bucket}.>"],
-            max_msgs_per_subject=config.history,
-            max_bytes=config.max_bytes,
-            max_age=config.ttl,
-            max_msg_size=config.max_value_size,
-            storage=config.storage,
-            num_replicas=config.replicas,
+            allow_direct=config.direct,
             allow_rollup_hdrs=True,
             deny_delete=True,
+            discard=api.DiscardPolicy.NEW,
+            duplicate_window=duplicate_window,
+            max_age=config.ttl,
+            max_bytes=config.max_bytes,
+            max_consumers=-1,
+            max_msg_size=config.max_value_size,
+            max_msgs=-1,
+            max_msgs_per_subject=config.history,
+            num_replicas=config.replicas,
+            storage=config.storage,
+            republish=config.republish,
         )
-        await self.add_stream(stream)
-
+        si = await self.add_stream(stream)
         assert stream.name is not None
+
         return KeyValue(
             name=config.bucket,
             stream=stream.name,
             pre=KV_PRE_TEMPLATE.format(bucket=config.bucket),
             js=self,
+            direct=si.config.allow_direct
         )
 
     async def delete_key_value(self, bucket: str) -> bool:
